import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy as sp
from scipy import stats
import warnings
from decimal import Decimal
import datetime
import logging


def get_logger(pathname):
    """
    Create log file similar to Ilaps-GUI log file

    Parameters
    ----------
    pathname : str
    Path to a log file(.txt). If file doesn't exists it is created.
    """
    logging.basicConfig(level=logging.INFO, format='%(levelname)s (%(name)s): %(asctime)s >>> %(message)s',
                        datefmt='%d-%b-%y %H:%M:%S', filename=pathname)
    logger = logging.getLogger("main")
    return logger


def get_timestamp(strTime):
    """format string time from iolite to timestamp"""
    return datetime.datetime.strptime(strTime, '%Y-%m-%d %H:%M:%S.%f')


def get_difference(start, now):
    """return time in seconds between 2 timestamps"""
    diff = now - start
    return diff.total_seconds()


def get_index(data, time):
    """return closest index of MS time given time in seconds"""
    for i in range(len(data.index)-1):
        if (data.index[i] <= time) and (data.index[i+1] > time):
            return i


def get_diff_lst(iolite):
    """return list of times in seconds from start to every start and end of laser ablation for spots"""
    lst = []
    for i in range(1, len(iolite['Timestamp'])-1):
        if (i-4) % 5 == 0:
            lst.append(get_difference(get_timestamp(
                iolite.loc[i-2, 'Timestamp']), get_timestamp(iolite.loc[i, 'Timestamp'])))
            lst.append(get_difference(get_timestamp(
                iolite.loc[i, 'Timestamp']), get_timestamp(iolite.loc[i+1, 'Timestamp'])))
    lst.append(get_difference(get_timestamp(
        iolite.loc[i, 'Timestamp']), get_timestamp(iolite.loc[i+1, 'Timestamp'])))
    return lst


def get_diff_lst_line(iolite):
    """return list of times in seconds from start to every start and end of laser ablation for lines"""
    lst = []
    for i in range(1, len(iolite['Timestamp'])-1):
        if (i-6) % 7 == 0:
            lst.append(get_difference(get_timestamp(
                iolite.loc[i-2, 'Timestamp']), get_timestamp(iolite.loc[i, 'Timestamp'])))
            lst.append(get_difference(get_timestamp(
                iolite.loc[i, 'Timestamp']), get_timestamp(iolite.loc[i+1, 'Timestamp'])))
    lst.append(get_difference(get_timestamp(
        iolite.loc[i-2, 'Timestamp']), get_timestamp(iolite.loc[i, 'Timestamp'])))
    return lst


def element_strip(elem):
    """formatting of element name to match colnames of reference material"""
    elem = elem.replace('(LR)', '').replace('(MR)', '').replace('(HR)', '')
    elem = ''.join([c for c in elem if c.isalpha()])
    return elem


def elem_resolution(elem):
    """formatting of element name to remove resolution for ELEMENT2 instrument"""
    elem = elem.replace('(LR)', '').replace('(MR)', '').replace('(HR)', '')
    return elem


def remove_outliers(data, offset, width):
    """
    Function to filter data by percentile value.

    Parameters
    ----------
    data: nparray
        array of data
    offset: float
        1-offset = upper treshold for percentile filtering. 
        Accepts values between 0 and 1.
    width: float
        Width od returned values, where 1-offset-width = lower treshold for percentile filtering.
        Accepts values between 0 and 1.

    Returns
    -------
    data_out : nparray
        Filtered data.
    """

    if offset < 0 or 1 < offset:
        raise ValueError(f'offset={offset}: Offset must be between 0 and 1')

    if width < 0 or 1 < width:
        raise ValueError(f'Width={width}: Width must be between 0 and 1')

    upper, lower = np.percentile(data, [round(
        100*(1-offset), 4), round(100*(1-offset-width), 4)])  # round to clear float error

    data_out = np.array([x for x in data if x >= lower and x <= upper])

    return data_out


def plot_data(data, isotopes=None, ax=None, *args, **kwargs):
    """
    Create a plot of MS time dependant data.

    Parameters
    ----------
    data : DataFrame
        Dataframe of ms data where column names are names of measured isotopes and index is time.
    isotopes : list
        List of isotopes to plot. (Optional)
    ax : matplotlib axes
        Matplotlib axes to show the plot. (Optional)

    """
    if ax is None:
        fig, ax = plt.subplots()
    if isotopes is not None:
        ax.plot(data[isotopes], *args, **kwargs)
    else:
        ax.plot(data, *args, **kwargs)


def correction(data, elem, internal_std):
    """
    Calculates internal standard correction.

    Parameters
    ----------
    data: DataFrame
        DataFrame with quantified values where columns are measured isotopes
    el: str
        Element used as internal standard
    internal_std: DataFrame
        DataFrame of values of internal standard  where columns are elements for 
        correction and each row represents one measurement e.g. Spot.
    """
    ratio = data[element_formater(elem, data.columns)].div(
        list(internal_std[elem]))
    return data.apply(lambda x: x/ratio)


def element_formater(elem, lst_of_elems):
    """matches the given element format to the one used in list"""
    if elem in lst_of_elems:
        return elem
    elif elem not in lst_of_elems:
        elem = elem.replace(' oxide', '')
        if elem in lst_of_elems:
            return elem
        elif elem not in lst_of_elems:
            elem = elem.replace('(LR)', '').replace(
                '(MR)', '').replace('(HR)', '')
            if elem in lst_of_elems:
                return elem
            elif elem not in lst_of_elems:
                elem = elem + '(LR)'
                if elem in lst_of_elems:
                    return elem
                elif elem not in lst_of_elems:
                    elem = elem.replace('(LR)', '')
                    elem = elem + '(MR)'
                    if elem in lst_of_elems:
                        return elem
                    elif elem not in lst_of_elems:
                        elem = elem.replace('(MR)', '')
                        elem = elem + '(HR)'
                        if elem in lst_of_elems:
                            return elem
                        elif elem not in lst_of_elems:
                            elem = elem.replace('(HR)', '')
                            elem = ''.join([c for c in elem if c.isalpha()])
                            if elem in lst_of_elems:
                                return elem
                            else:
                                return


def report(x, LoD, elem):
    """
    Replace values lower than limit of detection.
    If the value is above LoD, round to specific decimal place.
    """
    if x < LoD[elem]:
        return '< LoD'
    else:
        if x >= 100:
            return float(Decimal(x).quantize(Decimal('1')))
        else:
            if x >= 10:
                return float(Decimal(x).quantize(Decimal('0.1')))
            else:
                if x >= 1:
                    return float(Decimal(x).quantize(Decimal('0.01')))
                else:
                    if x >= 0:
                        return float(Decimal(x).quantize(Decimal('0.001')))


ox_name = {'Na23': 'Na2O (%)',
           'Mg24': 'MgO (%)',
           'Al27': 'Al2O3 (%)',
           'Si28': 'SiO2 (%)',
           'P31': 'P2O5 (%)',
           'K39': 'K2O (%)',
           'Ca44': 'CaO (%)',
           'Ti47': 'TiO2',
           'Mn55': 'MnO (%)',
           'Fe56': 'Fe2O3 (%)',
           'Co59': 'CoO',
           'Cu63': 'CuO (%)',
           'Sn118': 'SnO2',
           'Sb121': 'Sb2O3',
           'Pb208': 'PbO'}

ox_names = {el + ' oxide': ox for el, ox in ox_name.items()}

cols_to_perc = [x for x in ox_names.values() if '%' in x]


def z_score_method(df, variable_name, threshold=3):
    # Takes two parameters: dataframe & variable of interest as string
    columns = df.columns
    z = np.abs(sp.stats.zscore(df))

    outlier = []
    index = 0
    for item in range(len(columns)):
        if columns[item] == variable_name:
            index = item
    for i, v in enumerate(z[:, index]):
        if v > threshold:
            outlier.append(i)
        else:
            continue
    return outlier


def multivariate_outliers(df, cols=cols_to_perc, threshold=3):
    outliers = []
    for column in cols:
        outliers.extend(z_score_method(df, column, threshold))

    out_dict = {i: outliers.count(i) for i in range(len(df))}
    multi_out = [key for key, val in out_dict.items() if val > 7]

    return multi_out


def formatted_export(frame, path):
    writer = pd.ExcelWriter(path, engine='xlsxwriter')

    # Exporting analysis with outliers marked by red colour

    outliers = {}

    frame.replace('< LoD', np.nan, inplace=True)
    grouped = frame.groupby(frame.index)

    outliers_list = []
    for sample, group in grouped:
        if len(group) < 3:
            continue
        outliers[sample] = multivariate_outliers(group, threshold=1)

    frame.replace(np.nan, '< LoD', inplace=True)
    frame.to_excel(writer, sheet_name='Outliers')

    # formatting of excel
    workbook = writer.book
    worksheet = writer.sheets['Outliers']

    outlier_fmt = workbook.add_format({'color': 'red'})

    counter = 0
    tmp = ''
    for i, (sample, row) in enumerate(frame.iterrows()):
        if sample not in outliers.keys():
            continue

        if sample == tmp:
            counter += 1
        else:
            counter = 0
            tmp = sample
            idxs = outliers[sample]

        if counter in idxs:
            worksheet.set_row(row=i+1, cell_format=outlier_fmt)

    # exporting stats without outliers

    frame.replace('< LoD', np.nan, inplace=True)
    grouped = frame.groupby(frame.index)
    sumary = pd.DataFrame(columns=pd.concat(
        [pd.Series(['sample', 'stat']), pd.Series(frame.columns)]))
    if 'LoD' in frame.index:
        sumary = sumary.append(pd.concat(
            [pd.Series({'sample': 'LoD', 'stat': '-'}), frame.loc['LoD']]), ignore_index=True)

    # calculating stats
    for name, group in grouped:
        if name == 'LoD':
            continue

        # remove outliers
        # if name in outliers.keys():
        #     group.reset_index(inplace=True)
        #     group.drop(outliers[name], inplace=True)
        #     group.index = group['index']
        #     group.drop(['index'], axis=1, inplace=True)

        priemer = pd.concat(
            [pd.Series({'sample': name, 'stat': 'mean'}), group.mean()])
        sd = pd.concat(
            [pd.Series({'sample': name, 'stat': '2*std'}), group.std()*2])
        maxim = pd.concat(
            [pd.Series({'sample': name, 'stat': 'max'}), group.max()])
        minim = pd.concat(
            [pd.Series({'sample': name, 'stat': 'min'}), group.min()])
        med = pd.concat(
            [pd.Series({'sample': name, 'stat': 'median'}), group.median()])

        sumary = pd.concat(
            [sumary, pd.concat([priemer, sd, maxim, minim, med], axis=1).transpose()])

    sumary.replace(np.nan, '< LoD', inplace=True)

    sumary.to_excel(writer, sheet_name='Stats', index=False)

    # formatting of excel
    workbook = writer.book
    worksheet = writer.sheets['Stats']

    lod_fmt = workbook.add_format({'bold': True})
    bold_fmt = [workbook.add_format({'bg_color': '#E4FAEE', 'top': 1, 'bold': True}),
                workbook.add_format({'bg_color': '#FAF3E4', 'top': 1, 'bold': True})]
    normal_fmt = [workbook.add_format({'bg_color': '#E4FAEE', 'bold': False}),
                  workbook.add_format({'bg_color': '#FAF3E4', 'bold': False})]

    worksheet.set_row(row=1, cell_format=lod_fmt)

    tmp = False

    for i, (label, s) in enumerate(zip(sumary.index, sumary['sample'])):

        if s == 'LoD':
            continue

        if label == 0:
            worksheet.set_row(row=i+1, cell_format=bold_fmt[int(tmp)])
        else:
            worksheet.set_row(row=i+1, cell_format=normal_fmt[int(tmp)])

        if label == 4:
            tmp = not tmp

    writer.save()



from imgMS import MSData
from imgMS import MSEval
from imgMS.side_functions import *

import numpy as np
import xlsxwriter
import io
import os

from matplotlib.widgets import LassoSelector
from matplotlib.path import Path
import matplotlib.pyplot as plt

import itertools

stats_dict = {'mean':None, 'std':None, 'sum':None, 'min':None, 'max':None, 'med':None,}

class InteractiveAverage():
    """
    Creates interactive graph for the selection of the area of interest and
    calculates basic stats as mean, std, median, min and max. The values can
    be printed or saved as an excel file.  

    Parameters
    ----------
    msdata: MSData object from imgMS
        Data of elemental imaging where stats will be calculated from a selected area of the elemental distribution image. 
    """
    def __init__(self, msdata, elem=None):
        
        if elem is None: 
            self.elem = msdata.isotope_names[0]
        else: 
            if elem in msdata.isotope_names:
                self.elem = elem
            else:
                print(f'{elem} not in isotopes')
        self.msdata = msdata
        self.elmap = msdata.isotopes[self.elem].elmap
        self.get_attr()
        self.stats_dict = { e:{'mean':None, 'std':None, 'sum':None, 'min':None, 'max':None, 'med':None,} for e in self.msdata.isotope_names }
        

    def get_attr(self):
        """
        get all the data required for interactive mean calculation for given map
        """
        self.matrix = self.elmap.matrix
        self.shape = self.matrix.shape
        self.dx = self.elmap.dx
        self.dy = self.elmap.dy
        
        self.pixx = np.arange(self.shape[1])
        self.pixy = np.arange(self.shape[0])
        self.xv, self.yv = np.meshgrid(self.pixx,self.pixy)
        self.pix = np.vstack( (self.xv.flatten(), self.yv.flatten()) ).T
        print('pix', self.pix)

        self.array = np.array(self.matrix)
        self.array[np.isnan(self.array)] = 0
        
    def switch_elem(self, elem):
        """
        Change the element for interactive calculation
        
        Parameters
        ----------
        elem: str
        Name of the element, which will be used for selecting the boundaries of area where stats are calculated.
        """
        self.elem = elem
        self.elmap = self.msdata.isotopes[self.elem].elmap
        self.get_attr()
        
    def updateArray(self, array, indices):
        lin = np.arange(array.size)
        newArray = np.array(array.flatten())
        newArray[lin[~indices]] = 0.25
        newArray[lin[indices]] = 1
        return newArray.reshape(array.shape)
    
    def stats(self, indices):
        """
        Calculates the statistical values (mean, std, sum, minimum, 
        maximum and median) from selected area of image.
        
        Parameters
        ----------
        indices: list
        List of bool to select which pixels are in the area.
        """
        for e in self.msdata.isotopes:
            array = np.array(self.msdata.isotopes[e].elmap.matrix)
            lin = np.arange(array.size)
            newArray = np.array(array.flatten())
            
            mean = newArray[lin[indices]].mean()
            std = newArray[lin[indices]].std()
            suma = newArray[lin[indices]].sum()
            mina = newArray[lin[indices]].min()
            maxa = newArray[lin[indices]].max()
            meda = np.median(newArray[lin[indices]])
            
            self.stats_dict[e]['mean'] = newArray[lin[indices]].mean()
            self.stats_dict[e]['std'] = newArray[lin[indices]].std()
            self.stats_dict[e]['sum'] = newArray[lin[indices]].sum()
            self.stats_dict[e]['min'] = newArray[lin[indices]].min()
            self.stats_dict[e]['max'] = newArray[lin[indices]].max()
            self.stats_dict[e]['med'] = np.median(newArray[lin[indices]])
            
            print(e)
            print( f'mean: {mean}')
            print( f'std: {std}')
            print( f'sum: {suma}')
            print( f'min: {mina}')
            print( f'max: {maxa}')
            print( f'med: {meda}')

    def onSelect(self, verts):
        """
        Selection of area in the image.
        """
        p = Path(verts)
        ind = p.contains_points(self.pix, radius=1)
        alpha = self.updateArray(self.array, ind)
        self.im.set_alpha(alpha)
        self.fig.canvas.draw_idle()
        self.stats(ind)

    def __call__(self, ax=None, vmax=None):
        self.ax = ax
        if self.ax == None:
            self.fig, self.ax = plt.subplots()

        self.im = self.ax.imshow(self.array, vmax=vmax)
        self.ax.set_xlim([0, self.shape[1]])
        self.ax.set_ylim([0, self.shape[0]])

        lsso = LassoSelector(ax=self.ax, onselect=self.onSelect)

        plt.show()

    def export_stats(self, filename, sheetname='Sheet1'):
        """
        Export calculated stat for every isotope in the selected area 
        into excel.
        
        Parameters
        ----------
        filename: str
        Excel file name to save data.
        """
        workbook = xlsxwriter.Workbook(filename)
            
        wks1=workbook.add_worksheet(sheetname)
        
        wks1.write(0,0,'elem')
        wks1.write(1,0,'mean')
        wks1.write(2,0,'std')
        wks1.write(3,0,'sum')
        wks1.write(4,0,'min')
        wks1.write(5,0,'max')
        wks1.write(6,0,'med')
        
        for i, e in enumerate(self.msdata.isotopes):
            
            wks1.write(0,i+1,e)
            wks1.write(1,i+1,self.stats_dict[e]['mean'])
            wks1.write(2,i+1,self.stats_dict[e]['std'])
            wks1.write(3,i+1,self.stats_dict[e]['sum'])
            wks1.write(4,i+1,self.stats_dict[e]['min'])
            wks1.write(5,i+1,self.stats_dict[e]['max'])
            wks1.write(6,i+1,self.stats_dict[e]['med'])

        imgdata=io.BytesIO()
        self.fig.savefig(imgdata, format='png')
        wks1.insert_image(8,3, '', {'image_data': imgdata})

        workbook.close()
        

if __name__ == "__main__":
        
    '''
    import os

    files = os.listdir('../../Ilaps-v2/InteractiveMean/Data')
    files = [f for f in files if f.endswith('.xlsx')]

    print (f'proccesing files: {files}')

    for file in files:
    '''
        

    
    reader = MSData.DataReader(filename='/Users/nika/Library/CloudStorage/OneDrive-MUNI/Geologie/apatit_slobodnik/190415/mapa1_data.csv', filetype='csv', instrument='raw')
    data = MSData.MSData(reader)

    iolite = MSEval.Iolite('/Users/nika/Library/CloudStorage/OneDrive-MUNI/Geologie/apatit_slobodnik/190415/mapa1.Iolite.csv')

    data.select('iolite', s=10, iolite=iolite)

    data.create_maps(bcgcor_method='beginning')

    #data.isotopes['Ca44'].elmap()
    #plt.show()

    #mapa = data.isotopes['Ca44'].elmap
    
    lasso = InteractiveAverage(data, 'Ca44')
    lasso()
    lasso.export_stats('test_Ca.xlsx')
    
    print('=====================================')
    
    matrices = pd.ExcelFile('../../Ilaps-v2/InteractiveMean/Data/mapa1quant.xlsx') 
    d = MSData.MSData()
    d.isotope_names = matrices.sheet_names
    for el in d.isotope_names:
        d.isotopes[el] = MSData.Isotope(el)
        d.isotopes[el].elmap = MSData.ElementalMap()
    d.import_matrices(matrices)
    
    IA = InteractiveAverage(d)
    IA()
    
    
    
    
       

        






















        
